﻿using System;
using System.Collections.Generic;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.SimulationCore.Samples {
  [Item("Game of Life (discrete event)", "A discrete event based implementation of the game of life simulation.")]
  [Creatable("Simulations")]
  public class GameOfLifeDiscreteEventSimulation : DiscreteEventSimulation<GameOfLifeModel> {
    private GameOfLifeScenario Scenario {
      get { return Problem as GameOfLifeScenario; }
      set { Problem = value; }
    }

    private GameOfLifeDiscreteEventSimulation(GameOfLifeDiscreteEventSimulation original, Cloner cloner)
      : base(original, cloner) { }
    public GameOfLifeDiscreteEventSimulation() {
      Scenario = new GameOfLifeScenario();
      InitialAction = new InitializeGameOfLifeBoardAction() { AliveRate = Scenario.AliveRateParameter.Value.Value };
      Activities = new IActivity<GameOfLifeModel>[] { new CurrentTimeMonitor() };
      Reporters = new IReporter<GameOfLifeModel>[] { new GameOfLifeStateReporter() };
      Initialize();
    }

    private void Initialize() {
      Scenario.WidthParameter.ValueChanged += ValueChanged;
      if (Scenario.WidthParameter.Value != null) Scenario.WidthParameter.Value.ValueChanged += ParametersChanged;
      Scenario.HeightParameter.ValueChanged += ValueChanged;
      if (Scenario.HeightParameter.Value != null) Scenario.HeightParameter.Value.ValueChanged += ParametersChanged;
      Scenario.AliveRateParameter.ValueChanged += ValueChanged;
      if (Scenario.AliveRateParameter.Value != null) Scenario.AliveRateParameter.Value.ValueChanged += ParametersChanged;
    }


    private void ValueChanged(object sender, EventArgs e) {
      if (Scenario.WidthParameter.Value != null) Scenario.WidthParameter.Value.ValueChanged += ParametersChanged;
      if (Scenario.HeightParameter.Value != null) Scenario.HeightParameter.Value.ValueChanged += ParametersChanged;
      if (Scenario.AliveRateParameter.Value != null) Scenario.AliveRateParameter.Value.ValueChanged += ParametersChanged;
      ParametersChanged(sender, e);
    }

    private void ParametersChanged(object sender, EventArgs e) {
      Model = new GameOfLifeModel(Scenario.WidthParameter.Value.Value, Scenario.HeightParameter.Value.Value);
      ((InitializeGameOfLifeBoardAction)InitialAction).AliveRate = Scenario.AliveRateParameter.Value.Value;
    }

    protected override void OnPrepared() {
      Model = new GameOfLifeModel(Scenario.WidthParameter.Value.Value, Scenario.HeightParameter.Value.Value);
      base.OnPrepared();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GameOfLifeDiscreteEventSimulation(this, cloner);
    }
  }

  [StorableClass]
  public sealed class GameOfLifeModel : IModel {
    [Storable]
    public double CurrentTime { get; set; }
    [Storable]
    public bool[,] Grid { get; set; }

    private GameOfLifeModel(GameOfLifeModel original, Cloner cloner) {
      cloner.RegisterClonedObject(original, this);
      CurrentTime = original.CurrentTime;
      Grid = (bool[,])original.Grid.Clone();
    }
    public GameOfLifeModel(int width, int height) {
      Grid = new bool[height, width];
    }

    public object Clone() { return Clone(new Cloner()); }
    public IDeepCloneable Clone(Cloner cloner) {
      return new GameOfLifeModel(this, cloner);
    }
  }

  [StorableClass]
  sealed class CurrentTimeMonitor : Activity<GameOfLifeModel> {
    public override IEnumerable<Guid> MonitoredActionIds {
      get { return new[] { InitializeGameOfLifeBoardAction.Id, ModelTimeUpdateAction<GameOfLifeModel>.Id }; }
    }

    [StorableConstructor]
    private CurrentTimeMonitor(bool deserializing) : base(deserializing) { }
    private CurrentTimeMonitor(CurrentTimeMonitor original, Cloner cloner) : base(original, cloner) { }
    public CurrentTimeMonitor() : base(new SortedListEventQueue<GameOfLifeModel>()) { }

    public override void ManageEvents(GameOfLifeModel model, IAction<GameOfLifeModel> lastAction) {
      EventQueue.Push(new Event<GameOfLifeModel>(model.CurrentTime + 1, new UpdateGameOfLifeBoardAction()));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CurrentTimeMonitor(this, cloner);
    }
  }

  [StorableClass]
  sealed class GameOfLifeStateReporter : Reporter<GameOfLifeModel> {

    [StorableConstructor]
    private GameOfLifeStateReporter(bool deserializing) : base(deserializing) { }
    private GameOfLifeStateReporter(GameOfLifeStateReporter original, Cloner cloner) : base(original, cloner) { }
    public GameOfLifeStateReporter() { }

    public override void UpdateReporting(GameOfLifeModel model, ResultCollection results) {
      var heatmap = new HeatMap(model.Grid.GetLength(0), model.Grid.GetLength(1));
      for (var i = 0; i < heatmap.Rows; i++)
        for (var j = 0; j < heatmap.Columns; j++)
          heatmap[i, j] = model.Grid[i, j] ? 1 : 0;

      if (!results.ContainsKey("Visualization"))
        results.Add(new Result("Visualization", heatmap));
      else results["Visualization"].Value = heatmap;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GameOfLifeStateReporter(this, cloner);
    }
  }

  [StorableClass]
  sealed class InitializeGameOfLifeBoardAction : Action<GameOfLifeModel> {
    public static readonly Guid Id = new Guid("a6d37f71-ccef-4c2a-b11b-793fda0f9f4f");
    public override Guid ActionId {
      get { return Id; }
    }

    [Storable]
    public double AliveRate { get; set; }

    [StorableConstructor]
    private InitializeGameOfLifeBoardAction(bool deserializing) : base(deserializing) { }
    private InitializeGameOfLifeBoardAction(InitializeGameOfLifeBoardAction original, Cloner cloner)
      : base(original, cloner) {
      AliveRate = original.AliveRate;
    }
    public InitializeGameOfLifeBoardAction() {
      AliveRate = 0.2;
    }

    public override void Execute(GameOfLifeModel model) {
      var random = new Random();
      var rows = model.Grid.GetLength(0);
      var cols = model.Grid.GetLength(1);
      for (var i = 0; i < rows; i++)
        for (var j = 0; j < cols; j++) {
          if (random.NextDouble() < AliveRate) model.Grid[i, j] = true;
        }
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new InitializeGameOfLifeBoardAction(this, cloner);
    }
  }

  [StorableClass]
  sealed class UpdateGameOfLifeBoardAction : Action<GameOfLifeModel> {
    public static readonly Guid Id = new Guid("323bb2ef-bc66-488f-ac88-798b91565e62");
    public override Guid ActionId {
      get { return Id; }
    }

    [StorableConstructor]
    private UpdateGameOfLifeBoardAction(bool deserializing) : base(deserializing) { }
    private UpdateGameOfLifeBoardAction(UpdateGameOfLifeBoardAction original, Cloner cloner) : base(original, cloner) { }
    public UpdateGameOfLifeBoardAction() { }

    public override void Execute(GameOfLifeModel model) {
      var rows = model.Grid.GetLength(0);
      var cols = model.Grid.GetLength(1);
      var newGrid = new bool[rows, cols];
      for (var i = 0; i < rows; i++)
        for (var j = 0; j < cols; j++) {
          var neighbors = 0;
          if (i > 0 && j > 0 && model.Grid[i - 1, j - 1]) neighbors++;
          if (i > 0 && model.Grid[i - 1, j]) neighbors++;
          if (i > 0 && j < cols - 1 && model.Grid[i - 1, j + 1]) neighbors++;
          if (j > 0 && model.Grid[i, j - 1]) neighbors++;
          if (j < cols - 1 && model.Grid[i, j + 1]) neighbors++;
          if (i < rows - 1 && j > 0 && model.Grid[i + 1, j - 1]) neighbors++;
          if (i < rows - 1 && model.Grid[i + 1, j]) neighbors++;
          if (i < rows - 1 && j < cols - 1 && model.Grid[i + 1, j + 1]) neighbors++;
          if ((model.Grid[i, j] && neighbors == 2) || neighbors == 3) newGrid[i, j] = true;
        }
      model.Grid = newGrid;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new UpdateGameOfLifeBoardAction(this, cloner);
    }
  }
}
